/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_data.h"
#include "imaging_fdata.h"
#include "oirSupport.h"
#ifdef HP
#define finite isfinite
#endif
#define DUM 1

/**********************************************************
 * Usage:
 *    oirCompressSPData -data "inFiles" -mask maskFile -out outFile
 *       -ref refCoordFile
 *
 *    inFiles is a blank-separated list of input files
************************************************************** */
float * getSplineXCoord(imaging_detector_table_def *coord, int iRegion,
   int **coordIndex, int *nMon ){
   float *wcs ;
   int nx, ny, ix, iy;
   int n2, iCoord, jCoord, nCoord;
   float *coordData;
   int   *coordI;
   int reverse;

   wcs = imaging_detector_wcs(coord, iRegion);
   nx = coord->naxis[iRegion-1][0];
   ny = coord->naxis[iRegion-1][1];
   n2 = (int) nx/2;
   coordData  = (float *)malloc(nx*sizeof(float));
   coordI     = (int *)malloc(nx*sizeof(int));
   /* if coordinates decrease they need to be reversed */
   reverse = wcs[n2] < wcs[n2-1]; 
/*    search for central area where coordinates are monotonic */
   iCoord = n2;
   if (reverse) {
      while ((iCoord >= 0) && (wcs[iCoord] < wcs[iCoord-1]))
         --iCoord;
      jCoord = n2;
      while ((jCoord < nx-1) && (wcs[jCoord] < wcs[jCoord-1]))
         ++jCoord;
      nCoord = jCoord - iCoord + 1;
/*   
     flip both coordinates and indices and copy into output
*/
      for (ix=0; ix<nCoord; ix++) {
         coordData[ix]  = wcs[iCoord+nCoord-1-ix];
         coordI[ix]  = iCoord+nCoord-1-ix;
      }
   } else {  /* increasing values */
      while ((iCoord>= 0) && (wcs[iCoord] > wcs[iCoord-1]))
         --iCoord;
      jCoord = n2;
      while ((jCoord< nx-1) && (wcs[jCoord] > wcs[jCoord-1]))
         ++jCoord;
      nCoord= jCoord- iCoord+ 1;
      for (ix=0; ix<nCoord; ix++) {
         coordData[ix]   = wcs[ix+iCoord];
         coordI[ix]  = ix+iCoord;
      }
   }

   *nMon       = nCoord;
   *coordIndex = coordI;
   free(wcs);
   return coordData;

   } /* end of getSplineXCoord */

float * getSplineYCoord(imaging_detector_table_def *coord, int iRegion,
   float **newY) {
   float *wcs;
   int nx, ny, ix, iy;
   float *coordData;
   float *nyt;

   wcs = imaging_detector_wcs(coord, iRegion);
   nx = coord->naxis[iRegion-1][0];
   ny = coord->naxis[iRegion-1][1];
   coordData = (float *)malloc(nx*ny*sizeof(float));
   nyt       = (float *)malloc(ny*sizeof(float));
   for (ix=0; ix<nx; ix++) for (iy=0;iy<ny; iy++)
      coordData[iy+ix*ny] = wcs[ix + nx*(ny+iy)];
   for (iy=0; iy<ny; iy++) nyt[iy] = (float)(1+iy-coord->crpix[iRegion-1][1]);
   free(wcs);
   *newY = nyt;
   return coordData;
} /* end of getSplineYCoord */

void spRectifyY(int nx, int ny, short *inData, float *oldY, float *newY, 
   float *outData) {

   float *outWork, *inWork;
   int ix, iy;

   inWork = (float *)malloc(ny * sizeof(float));
   for (ix = 0; ix<nx; ix++) {
      for (iy=0;iy<ny; iy++)  inWork[iy] = (float)inData[ix+iy*nx]+32768.;
      outWork = minrtsSpline(ny, oldY+ix*ny, inWork, ny, newY, 0.1);
      for (iy=0;iy<ny; iy++)  outData[ix+nx*iy] = outWork[iy];
      free(outWork);
   }
} /* end of spRectifyY */

void spRectifyX(int nx, float *inData, float *xCoord,
   float *xRefCoord, int* xIndex, int *xRefIndex, int nMon, int nMonRef) {

   int i,j;
   float x;
   float *inWork, *outWork;
/*
      copy valid data montonically into inWork
      interpolate onto reference coordinate pixels
      and copy back
*/
   inWork = (float *)malloc(nMon * sizeof(float));
   for (i=0;i<nMon;i++) inWork[i] = inData[xIndex[i]];
   outWork = minrtsSpline(nMon, xCoord, inWork, nMonRef, xRefCoord, 0.1);
   for (i=0;i<nx;i++) inData[i] = 0.;
   for (i=0;i<nMonRef;i++) inData[xRefIndex[i]] = outWork[i];
   free (inWork);
   free (outWork);
   return;
} /* end of spRectifyX */

void Usage() {
   printf("Usage: oirCompressSPData -data ""inFiles"" \n");
   printf("   [-mask maskFile]\n");
   printf("-   out outFile\n");
   printf("   [-ref refCoordFile]\n");
   printf("   [-nDrop dropPoints]\n");
   printf("inFiles: a blank-separated list of input files\n");
   printf("enclosed in double quotes if indeed blank-separated\n");
   printf("MaskFile: the mask generated by oirSciPhotCal\n");
   printf("outFile: output file name, input to oirFormFringes\n");
   printf("refCoordFile: contains coordinate distortion info:\n");
   printf("defaults to 1st inFile\n");
}

int main (int argc, char** argv) {

   char **fileList;
   int nFiles, iFile;
   int nRegion, iRegion;
   int maxstep;
   int ndit;
   int i,j,k,ierr;
   long nRows, iRow, oRow;
   int nx, ny, ix, iy;
   int nxList[12], nyList[12];
   long ipix;
   short *inPix;
   float *maskImages[12];
   int isMask;
   int iKey;
   int iarg;

   char outFileName[130];
 
   fitsfile *inFile;
   imaging_data_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *inDetectorDef;
   imaging_detector_table_def *refDetectorDef;
   imaging_detector_table_def *outDetectorDef;
   imaging_data_row *inData;
   imaging_fdata_row *outData;

   float *splineXCoord[2];
   float *splineYCoord[2];
   float *splineData[2];
   float *newY[2];
   int   *xIndex[2];
   int    nMon[2];

   int oldTarget, isTarget;
   int nDrop;
   fitsfile *outFile  = NULL;
   char* inFileName   = NULL;
   char* maskFileName = NULL;
   char* refFileName  = NULL;
   char *bang = "!";
/*
      parse input
*/
   ierr = 0;
   iarg = 1;
   outFileName[0] = 0;
   nDrop = 2;
   while (iarg < argc) {
      if (strcmp(argv[iarg], "-data")==0) 
        inFileName = argv[++iarg];
      if (strcmp(argv[iarg], "-mask")==0) 
        maskFileName = argv[++iarg];
      if (strcmp(argv[iarg], "-nDrop")==0) 
        sscanf(argv[++iarg],"%i",&nDrop);
      if (strcmp(argv[iarg], "-out")==0) {
         strcpy(outFileName, bang);
         strcat(outFileName, argv[++iarg]);
      }
      if (strcmp(argv[iarg], "-ref")==0) 
        refFileName = argv[++iarg];
      ++iarg;
   }
   if ((inFileName == NULL) || outFileName[0] == 0) {
      Usage();
      return 1;
   }
/*
      get reference detector data
*/
   if (refFileName == NULL) refFileName = fileList[0];
      printf("   using %s as coord reference file\n", refFileName);
   fits_open_file (&inFile, refFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR: fits openfile of %s gave error %i \n",refFileName,
         ierr);
       printf("%s \n",strerror(errno));
      return 1;
   }
   refDetectorDef = read_imaging_detector_table(inFile, &ierr);
   i = fits_close_file(inFile, &ierr);
/*

/*        break up input file list and print what you find */
   fileList = oirStrSplit(inFileName, &nFiles);

   for (i=0; i<nFiles; i++) {printf("   %s \n", fileList[i]);}

   ierr = 0;
/*
      open first file and get detector info and some keywords; then close file
*/
   fits_open_file (&inFile, fileList[0], READONLY, &ierr);
   if (ierr != 0) {
      printf("ERROR:fits openfile of %s gave error %i \n",fileList[0],
         ierr);
       printf("%s \n",strerror(errno));
      return 1;
   }
   i = fits_read_key(inFile, TINT, "HIERARCH ESO DET NDIT", &ndit,
      0, &ierr);
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_data_table(inFile, &ierr);
   inData         = make_imaging_data_row(idata);
   nRegion        = idata->nregion;
   for (iRegion=0;iRegion<nRegion;iRegion++) {
      nxList[iRegion] = outDetectorDef->naxis[iRegion][0];
      nyList[iRegion] = outDetectorDef->naxis[iRegion][1];
   }
/*
      need maxstep for output table definition
*/
   iKey = table_find_key(idata->def, "MAXSTEP"); 
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep, 
      &ierr);
/*
      reset output data sizes to reflect compression.
      move coordinate information from reference row
      of reference file into output.
*/
   for (iRegion=0; iRegion<nRegion; iRegion++) {
#ifdef DUM
      outDetectorDef->naxis[iRegion][1] = 1;
      for (i=0;i<2;i++) {
         outDetectorDef->crval[iRegion][i] = refDetectorDef->crval[1][i];
         outDetectorDef->crpix[iRegion][i] = refDetectorDef->crpix[1][i];
         for (k=0;k<refDetectorDef->max_coef;k++) 
            outDetectorDef->dmc[iRegion][k][i] = refDetectorDef->dmc[1][k][i];
         for (j=0;j<2;j++) {
            outDetectorDef->cd[iRegion][i+2*j] = refDetectorDef->cd[1][i+2*j];
            for (k=0;k<refDetectorDef->max_coef;k++) 
               outDetectorDef->dmp[iRegion][k][i][j] = refDetectorDef->dmp[1][k][i][j];
         }
      }
#endif
   }

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      printf ("ERROR: opening fits output %s %i\n", outFileName, ierr);
      printf ("%s \n", strerror(errno), ierr);
      return 1;
      }
/*    
      copy main header from input to output and close input
*/
   i = fits_movabs_hdu(inFile, 1, 0, &ierr);
   i = fits_copy_header(inFile, outFile, &ierr);
   i = fits_close_file(inFile, &ierr);

   odata = make_imaging_fdata_table_def (outFile, 
      outDetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      printf ("ERROR: creating data definition\n");
      return 1;
      }
   outData = make_imaging_fdata_row(odata);
   oRow = 1;
/*
      get wcs info for central windows
      allocate space for interpolated data
*/
   for (iRegion=0; iRegion<2; iRegion++) {
      splineXCoord[iRegion] = getSplineXCoord(refDetectorDef, iRegion+2, 
         xIndex+iRegion, nMon+iRegion);
      splineYCoord[iRegion] = getSplineYCoord(refDetectorDef, iRegion+2, newY+iRegion);
   }
   splineData[0] = (float *)malloc(nxList[1]*nyList[1]*sizeof(float));
   splineData[1] = (float *)malloc(nxList[2]*nyList[2]*sizeof(float));

/*
      get mask data
*/
   ierr = 0;
   if (NULL == maskFileName) isMask = (1==0); 
   else {
      oirGetMaskImages(maskFileName, maskImages, nRegion, nxList, nyList, 
         &ierr);
      isMask = (ierr == 0);
   }
   if (isMask) printf("   using maskfile %s \n", maskFileName); 
   else if (ierr !=0) {
      printf("ERROR: maskfile %s gave error %i\n",maskFileName, ierr);
      return 1;
   }
/*   
    loop over input files
*/
   oldTarget = 0; /* start looking for Target frames */
   for (iFile=0; iFile<nFiles; iFile++) {
      fits_open_file (&inFile, fileList[iFile], READONLY, &ierr);
      if (ierr != 0) {
         printf("ERROR:fits openfile of %s gave error %i\n",fileList[iFile],
            ierr);
         printf("%s \n",strerror(errno));
         return 1;
      }
      inDetectorDef = read_imaging_detector_table(inFile, &ierr);
      idata = open_imaging_data_table(inFile, &ierr);
      i = fits_get_num_rows (inFile,  &nRows, &ierr); 
/* suppress autoscaling to avoid overflow */
      for (i=0; i<idata->nregion; i++) 
         fits_set_tscale(inFile, 1+idata->data_col[i], 1.0, 0.0, &ierr);
/*
         loop over input rows
*/
      for (iRow=1; iRow<=nRows; iRow++) {
/*
         check if chopping is on Target
         otherwise skip row
*/
         i = read_imaging_data_table(inFile, idata, inData, iRow, &ierr);
         isTarget = inData->targtype[1] == 'T';
         if (isTarget) ++oldTarget; else oldTarget = 0;
/*
            check for invalid time; this indicates EXPREP failure and
            the OPD values are also invalid.  If time and
            Chop OK, start processing
*/
         if ((finite(inData->time))&& (oldTarget>nDrop)) {
/*
            copy nonData info
*/
            outData->time = inData->time;
            outData->opd[0] = inData->opd[0];
            outData->opd[1] = inData->opd[1];
            outData->local_opd[1] = inData->local_opd[1];
            outData->local_opd[0] = inData->local_opd[0];
            outData->exp = inData->exp;
            outData->offset[0] = inData->offset[0];
            outData->offset[1] = inData->offset[1];
            outData->rotation  = inData->rotation;
            outData->frame     = inData->frame;
            outData->step      = inData->step;
            outData->reference = inData->reference;
            for (i=0; i<idata->nregion; i++) 
               outData->targtype[i] = inData->targtype[i];  
/*
             rectify interferometric channels in Y-direction then compress with 
             masking.  Ignore channels 1 and 4;
*/
            nx = inDetectorDef->naxis[1][0];
            ny = inDetectorDef->naxis[1][1];

            for (iRegion=1; iRegion<3; iRegion++) {
               spRectifyY(nx,ny,(short *)inData->data[iRegion],
                  splineYCoord[iRegion-1],newY[iRegion-1], splineData[iRegion-1]);
               for (ix=0; ix<nx; ix++) {
#ifdef DUM
                  outData->data[iRegion][ix] = 0.;
                  ipix = ix;
                  if (isMask) for (iy=0;iy<ny;iy++) {
                     outData->data[iRegion][ix] +=
                        splineData[iRegion-1][ipix]*maskImages[iRegion][ipix];
                     ipix += nx;
                  } else for (iy=0;iy<ny;iy++) {
                     outData->data[iRegion][ix] +=
                        splineData[iRegion-1][ipix];
                     ipix += nx;
                  }
#endif
#ifndef DUM
for (iy=0;iy<ny;iy++) outData->data[iRegion][ix+nx*iy] = splineData[iRegion-1][ix+nx*iy];
#endif
               } /* loop over ix */
/*
                    interpolate xarrays to reference frequency grid
*/
               spRectifyX(nx, outData->data[iRegion], splineXCoord[iRegion-1],
                  splineXCoord[0], xIndex[iRegion-1], xIndex[0],
                  nMon[iRegion-1], nMon[0] );
            } /* loop over regions */
/*
            write output row; 
*/
            i = write_imaging_fdata_table(outFile, odata, outData, 
               oRow++, &ierr);
         }  /* inputdata is valid for this row */
      } /* loop over input rows in this file */
      i = fits_close_file(inFile, &ierr);
      kill_imaging_detector_table_def(inDetectorDef);
   } /* end of loop over iFile */
   i = close_imaging_fdata_table(outFile, odata, &ierr);
   kill_imaging_fdata_row(outData);
   kill_imaging_data_row(inData);
   kill_imaging_data_table_def (idata,1);
   kill_imaging_detector_table_def(refDetectorDef);
   for (iRegion=0;iRegion<nRegion;iRegion++) free(maskImages[iRegion]);
   for (iRegion=0; iRegion<2; iRegion++) {
      free(splineXCoord[iRegion]);
      free(splineYCoord[iRegion]);
      free(splineData[iRegion]);
      free(xIndex[iRegion]);
   }
   return 0;
}
